"""Helper methods for accessing single pixel from a rasterio file object.

"""
import rasterio
import rasterio.warp
import rasterio.crs

from typing import Union, Iterable, Optional, List, Tuple

RowCol = Tuple[int, int]
XY = Tuple[float, float]
LonLat = Tuple[float, float]
SomeCoord = Union[RowCol, XY, LonLat]
PixelValue = Union[float, int]


NOTSET = object()


def make_pixel_extractor(
    mode="pixel",
    band=1,
    src_nodata_fallback=None,
    src_nodata_override=None,
    dst_nodata=NOTSET,
):
    """Returns function that can extract single pixel from opened rasterio file.

    Signature of the returned function is:
        `src, coordinate_tuple, [band] -> pixel`

    Where coordinate_tuple is interpreted according to `mode`


    mode       - How to interpret coordinate:
                  - pixel: (row, col)
                  - native: (x, y) in file native coordinate space
                  - lonlat: (lon, lat)  (specifically EPSG:4326)

    band       - Default band to read, can be over-written on read

    dst_nodata - when set use that instead of defaulting to src nodata value,
                 can be set to `None` to remap to `None`

    src_nodata_fallback - nodata value to use if src file is missing nodata value

    src_nodata_override - when set use that instead of what's in the file,
                          useful when nodata metadata is incorrect in the file
                          but correct value is available out of band.

    """
    default_band = band

    if dst_nodata is NOTSET:

        def _dst_nodata(src_nodata):
            return src_nodata

    else:

        def _dst_nodata(src_nodata):
            return dst_nodata

    def remap_pix(pix, src_nodata, dst_nodata):
        # TODO: special case src_nodata is nan case
        return dst_nodata if pix == src_nodata else pix

    def extract_pixel(src, coord, band=default_band):
        ri, ci = coord

        src_nodata = _resolve_nodata(
            src, band, fallback=src_nodata_fallback, override=src_nodata_override
        )

        dst_nodata = _dst_nodata(src_nodata)

        if 0 <= ri < src.height and 0 <= ci < src.width:
            window = ((ri, ri + 1), (ci, ci + 1))

            pix = src.read(band, window=window)
            # TODO: support band being a list of bands
            return remap_pix(pix[0][0], src_nodata, dst_nodata)
        else:
            return dst_nodata

    def extract_native(src, coord, band=default_band):
        return extract_pixel(src, src.index(*coord), band=band)

    def extract_lonlat(src, coord, band=default_band):
        lon, lat = coord
        x, y = rasterio.warp.transform(
            rasterio.crs.CRS.from_epsg(4326), src.crs, [lon], [lat]
        )
        xy = (x[0], y[0])
        return extract_native(src, xy, band=band)

    extractors = dict(pixel=extract_pixel, native=extract_native, lonlat=extract_lonlat)

    extractor = extractors.get(mode)
    if extractor is None:
        raise ValueError("Only support mode=<pixel|native|lonlat>")

    return extractor


def _resolve_nodata(src, band, fallback=None, override=None):
    """Figure out what value to use for nodata given a band and fallback/override
    settings

    :param src: Rasterio file
    """
    if override is not None:
        return override

    band0 = band if isinstance(band, int) else band[0]
    nodata = src.nodatavals[band0 - 1]

    if nodata is None:
        return fallback

    return nodata


def _mode_value(
    pixel: Optional[RowCol] = None,
    xy: Optional[XY] = None,
    lonlat: Optional[LonLat] = None,
) -> Union[Tuple[str, SomeCoord], Tuple[None, None]]:
    if pixel is not None:
        return "pixel", pixel

    if xy is not None:
        return "native", xy

    if lonlat is not None:
        return "lonlat", lonlat

    return (None, None)


def read_pixels(
    urls: Iterable[str],
    pixel: Optional[RowCol] = None,
    xy: Optional[XY] = None,
    lonlat: Optional[LonLat] = None,
    band: int = 1,
    **kwargs
) -> List[PixelValue]:
    """Read a single pixel at the same location from a bunch of different files.

    Location can be specified in 3 different ways:

      pixel  (row: int, column: int)  -- in pixel coords
      xy     (X: float, Y: float)     -- in Projected coordinates of the native CRS of the image
      lonlat (lon: float, lat: float) -- in EPSG:4326
    """
    mode, coord = _mode_value(pixel=pixel, xy=xy, lonlat=lonlat)
    if mode is None:
        raise ValueError("Have to supply one of: pixel, xy, or lonlat.")

    extractor = make_pixel_extractor(mode=mode, band=band, **kwargs)

    def read_from_url(url):
        url = rasterio.parse_path(url)
        with rasterio.DatasetReader(url, sharing=False) as src:
            return extractor(src, coord=coord)

    return [read_from_url(url) for url in urls]



from typing import Dict, Optional, Any
import dask
import dask.array as da
import numpy as np
import xarray as xr
import numexpr as ne
import functools
from ._dask import randomize, flatten_kv, unflatten_kv


def apply_numexpr_np(
    expr: str,
    data: Optional[Dict[str, Any]] = None,
    dtype=None,
    out: Optional[np.ndarray] = None,
    casting="safe",
    order="K",
    **params,
) -> np.ndarray:
    """
    Apply numexpr to numpy arrays
    """

    if data is None:
        data = params
    else:
        data.update(params)

    if out is None and dtype is not None:
        # This needs to be np.ndarray
        arrays = [x for x in data.values() if isinstance(x, np.ndarray)]
        if len(arrays) == 0:
            raise ValueError("Could not find any arrays on input")

        sample_input = arrays[0]
        out = np.empty_like(sample_input, dtype=dtype)

    return ne.evaluate(expr, local_dict=data, out=out, casting=casting, order=order)


def apply_numexpr(
    expr: str,
    xx: xr.Dataset,
    dtype=None,
    name="result",
    casting="safe",
    order="K",
    **params,
):
    """
    Apply numexpr to variables within a Dataset.

    numexpr library offers a limited subset of types and operations supported
    by numpy, but is much faster and memory efficient, particularly for complex
    expressions. See numexpr documentation for a more detailed explanation of
    performance advantages of using this library over numpy operations,
    summary: single pass over input memory, no temporary arrays, cache
    locality.

    :param expr: Numexpr compatible string to evaluate
    :param xx: Dataset object that contains arrays to be used in the ``expr`` (can be Dask)
    :param dtype: specify output dtype
    :param name: Used to name computation when input is Dask
    :param casting: Passed to ``numexpr.evaluate``
    :param order: Passed to ``numexpr.evaluate``
    :param params: Any other constants you use in the expression
    :raturns: xr.DataArray containing result of the equation (Dask is input is Dask)

    Example:

    .. code-block:: python

       # Given a Dataset with bands `red` and `nir`
       xx = dc.load(..., measurements=["red", "nir"], dask_chunks={})

       # Compute NDVI (ignore nodata for simplicity of the example)
       ndvi = apply_numexpr("(_1f*nir - red)/(_1f*nir + red)",
                            xx,
                            dtype='float32',   # Output is float32
                            _1f=np.float32(1)  # Define constant `_1f` being a float32(1),
                                               # used for casting to float32
                           )
    """

    bands = {}
    sample_band = None

    for band, x in xx.data_vars.items():
        band = str(band)

        if band in params:
            raise ValueError(f"Variable: `{band}` is aliased by a parameter")
        if band in expr:
            bands[band] = x.data

            if sample_band is None:
                sample_band = x

    if sample_band is None:
        raise ValueError("Found no bands on input")

    op = functools.partial(
        apply_numexpr_np, expr, dtype=dtype, casting=casting, order=order, **params
    )

    if dask.is_dask_collection(xx):
        # Passing through dictionary of Dask Arrays didn't work, so we have
        # adaptor that accepts var args in the form of [k0,v0,  k1,v1, ...] and then reconstructs dict
        data = da.map_blocks(
            lambda op, *bands : op(unflatten_kv(bands)),
            op,
            *flatten_kv(bands),
            name=randomize(name),
            dtype=dtype,
        )
    else:
        data = op(bands)

    return xr.DataArray(
        data=data,
        attrs=sample_band.attrs,
        dims=sample_band.dims,
        coords=sample_band.coords,
        name=name,
    )


def safe_div(x1: xr.DataArray, x2: xr.DataArray, dtype='float32') -> xr.DataArray:
    """
    Compute ``x1.astype(dtype)/x2.astype(dtype)`` taking care of cases where x2==0.

    For every element compute the following:

    ::

      x2 is 0 => NaN
      else    => float(x1)/float(x2)

    TODO: currently doesn't treat nodata values in any special way.
    """
    dtype = np.dtype(dtype)

    # TODO: support nodata on input
    return apply_numexpr("where(x2 == 0, nan, (_1f * x1) / x2)",
                         xr.Dataset(dict(x1=x1, x2=x2)),
                         dtype=dtype,
                         nan=dtype.type("nan"),
                         _1f=dtype.type(1))
